# coding: utf-8
"""# `importnb` Jupyter magic extensions
"""

"""    %importnb --stdout --stderr --display --shell
"""

import argparse
import inspect
from functools import partial
from importlib import import_module

import pkg_resources


def get_module_object(str):
    module, object = str.split(":", 1)
    return getattr(import_module(module), object)


parser = argparse.ArgumentParser(
    description="""Define the importnb loader properties."""
)
parser.add_argument("--cls", type=get_module_object, default="importnb:Notebook")
parser.add_argument("--fuzzy", action="store_true")

try:
    from IPython.core import magic_arguments
    from IPython.core.magic import (
        Magics,
        magics_class,
        line_magic,
        cell_magic,
        line_cell_magic,
    )

    __IPYTHON__ = True
except ImportError:
    __IPYTHON__ = False


class ImportNbExtensionBase:
    loaders = None

    def __init__(self, shell, loader=None):
        self.loaders = []
        # A default loader to install
        if loader:
            self.loaders.append(loader(position=-1).__enter__())


if __IPYTHON__:

    @magics_class
    class ImportNbExtension(Magics, ImportNbExtensionBase):
        loaders = None

        def __init__(self, shell, loader=None):
            Magics.__init__(self, shell)
            ImportNbExtensionBase.__init__(self, shell, loader)

        @line_cell_magic
        def importnb(self, line, cell=None):
            if line.strip() == "pop":
                return self.pop()
                module.__package__

            details = vars(parser.parse_args(line.split()))
            self.loaders.append(details.pop("cls")(**details))

            if cell is None:
                self.loaders[-1].__enter__()
                return

            with self.loaders.pop(-1):
                self.parent.run_cell(cell)

        def unload(self):
            while self.loaders:
                self.pop()

        def pop(self):
            self.loaders.pop().__exit__(None, None, None)


else:

    class ImportNbExtension(ImportNbExtensionBase):
        ...


manager = None


def IPYTHON_MAIN():
    """Decide if the Ipython command line is running code."""
    import pkg_resources

    runner_frame = inspect.getouterframes(inspect.currentframe())[-2]
    return (
        getattr(runner_frame, "function", None)
        == pkg_resources.load_entry_point(
            "ipython", "console_scripts", "ipython"
        ).__name__
    )


def load_ipython_extension(ip=None):
    global manager, module
    if IPYTHON_MAIN():
        from .parameterize import Parameterize as Notebook
    else:
        from .loader import Notebook

    # Auto loading only works in IPython and
    # we only read need it when there are parameters.
    manager = ImportNbExtension(ip, Notebook)

    if ip:
        ip.register_magics(manager)
        ip.user_ns.update(__path__=[str(__import__("pathlib").Path().absolute())])
        from .completer import load_ipython_extension

        load_ipython_extension(ip)

        ip.user_ns["reload"] = __import__("importlib").reload


def unload_ipython_extension(ip=None):
    global manager
    ip and manager and manager.unload()


"""# Developer
"""

if __name__ == "__main__":
    from importnb.utils.export import export

    export("ipython_extension.ipynb", "../ipython_extension.py")
    # m = Notebook(shell=True).from_filename('extensions.ipynb')
    # print(__import__('doctest').testmod(m, verbose=2))



# coding: utf-8
"""# Parameterize

The parameterize loader allows notebooks to be used as functions and command line tools.  A `Parameterize` loader will convert an literal ast assigments to keyword arguments for the module.
"""

import argparse
import ast
import inspect
import sys
from copy import deepcopy
from functools import partial, partialmethod
from importlib.util import find_spec, spec_from_loader
from inspect import Parameter, Signature, signature
from pathlib import Path

from .loader import _GTE38, Notebook, module_from_spec

if _GTE38:
    from importlib._bootstrap import _load_unlocked
else:
    from importlib._bootstrap import _installed_safely


class FindReplace(ast.NodeTransformer):
    def __init__(self, globals, parser):
        self.globals = globals
        self.parser = parser
        self.argv = sys.argv[1:]
        self.parameters = []

    def visit_Assign(self, node):
        if len(node.targets) == 1 and isinstance(node.targets[0], ast.Name):
            target, parameter = node.targets[0].id, node.value
            try:
                parameter = ast.literal_eval(parameter)
            except:
                return node

            if target[0].lower():
                extras = {}
                if isinstance(parameter, bool):
                    extras.update(
                        action="store_" + ["true", "false"][parameter],
                        help="{} = {}".format(target, not parameter),
                    )
                else:
                    extras.update(
                        help="{} : {} = {}".format(
                            target, type(parameter).__name__, parameter
                        )
                    )
                try:
                    self.parser.add_argument(
                        "--%s" % target, default=parameter, **extras
                    )
                except argparse.ArgumentError:
                    ...
                self.parameters.append(
                    Parameter(target, Parameter.KEYWORD_ONLY, default=parameter)
                )
                if ("-h" not in self.argv) and ("--help" not in self.argv):
                    ns, self.argv = self.parser.parse_known_args(self.argv)
                    if target in self.globals:
                        node = ast.Expr(ast.Str("Skipped"))
                    elif getattr(ns, target) != parameter:
                        node.value = ast.parse(str(getattr(ns, target))).body[0].value
        return node

    @property
    def signature(self):
        return Signature(self.parameters)

    def visit_Module(self, node):
        node.body = list(map(self.visit, node.body))
        self.parser.description = ast.get_docstring(node)
        self.parser.parse_known_args(self.argv)  # run in case there is a help arugment
        return node

    def generic_visit(self, node):
        return node


def copy_(module):
    new = type(module)(module.__name__)
    return new.__dict__.update(**vars(module)) or new


class Parameterize(Notebook):
    __slots__ = Notebook.__slots__ + ("globals",)

    def __init__(
        self,
        fullname=None,
        path=None,
        *,
        lazy=False,
        fuzzy=True,
        markdown_docstring=True,
        position=0,
        globals=None,
        main=False,
        **_globals
    ):
        super().__init__(
            fullname, path, lazy=lazy, fuzzy=fuzzy, position=position, main=main
        )
        self.globals = globals or {}
        self.globals.update(**_globals)
        self._visitor = FindReplace(
            self.globals, argparse.ArgumentParser(prog=self.name)
        )

    def exec_module(self, module):
        self._visitor = FindReplace(self.globals, self._visitor.parser)
        module.__dict__.update(**self.globals)
        return super().exec_module(module)

    def visit(self, node):
        return super().visit(self._visitor.visit(node))

    @classmethod
    def load(cls, object, **globals):
        return parameterize(super().load(object), **globals)


"""    with Parameterize():
        reload(foo)

    with Parameterize(a=1234123):
        reload(foo)

    with Parameterize(a="🤘"):
        reload(foo)
"""

"""    import foo
"""


def parameterize(object, **globals):
    with Parameterize(**globals):
        if isinstance(object, str):
            object = module_from_spec(find_spec(object))

    object.__loader__ = Parameterize(
        object.__loader__.name, object.__loader__.path, **globals
    )

    def call(**parameters):
        nonlocal object, globals
        object = copy_(object)
        keywords = {}
        keywords.update(**globals), keywords.update(**parameters)
        if _GTE38:
            Parameterize(object.__name__, object.__file__, **keywords).exec_module(
                object
            )
        else:
            with _installed_safely(object):
                Parameterize(object.__name__, object.__file__, **keywords).exec_module(
                    object
                )
        return object

    object.__loader__.get_code(object.__name__)
    call.__doc__ = object.__doc__ or object.__loader__._visitor.parser.format_help()
    call.__signature__ = object.__loader__._visitor.signature
    return call



# coding: utf-8
import argparse
import os
from doctest import FAIL_FAST, OPTIONFLAGS_BY_NAME, testfile, testmod

from ..loader import Notebook


def _test():
    parser = argparse.ArgumentParser(description="doctest runner")
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="print very verbose output for all tests",
    )
    parser.add_argument(
        "-o",
        "--option",
        action="append",
        choices=OPTIONFLAGS_BY_NAME.keys(),
        default=[],
        help=(
            "specify a doctest option flag to apply"
            " to the test run; may be specified more"
            " than once to apply multiple options"
        ),
    )
    parser.add_argument(
        "-f",
        "--fail-fast",
        action="store_true",
        help=(
            "stop running tests after first failure (this"
            " is a shorthand for -o FAIL_FAST, and is"
            " in addition to any other -o options)"
        ),
    )
    parser.add_argument("file", nargs="+", help="file containing the tests to run")
    args = parser.parse_args()
    testfiles = args.file
    # Verbose used to be handled by the "inspect argv" magic in DocTestRunner,
    # but since we are using argparse we are passing it manually now.
    verbose = args.verbose
    options = 0
    for option in args.option:
        options |= OPTIONFLAGS_BY_NAME[option]
    if args.fail_fast:
        options |= FAIL_FAST
    for filename in testfiles:
        if any(map(filename.endswith, (".py", ".ipynb"))):
            # It is a module -- insert its dir into sys.path and try to
            # import it. If it is part of a package, that possibly
            # won't work because of package imports.
            failures, _ = testmod(
                Notebook.load(filename), verbose=verbose, optionflags=options
            )
        else:
            failures, _ = testfile(
                filename, module_relative=False, verbose=verbose, optionflags=options
            )
        if failures:
            return 1
    return 0


if __name__ == "__main__":
    _test()

if __name__ == "__main__":
    from .export import export

    export("nbdoctest.ipynb", "../../utils/nbdoctest.py")



import sys

from .parameterize import Parameterize

file = sys.argv[1] if len(sys.argv) > 1 else None

if file == __file__:
    sys.argv = [sys.argv[0]] + [sys.argv[2:]]

    file and Parameterize.load(file)



# coding: utf-8
"""# Decode `nbformat` with line numbers

`importnb` decodes notebooks with the `nbformat` in valid source code.

We consider three kinds of cells.
"""

import linecache
import textwrap
from functools import partial
from json import load as _load
from json import loads as _loads
from json.decoder import (
    WHITESPACE,
    WHITESPACE_STR,
    JSONDecoder,
    JSONObject,
    py_scanstring,
)
from json.scanner import py_make_scanner


"""Output the strings slice that the source came from.
"""


def scanstring(s, end, strict=True, **kwargs):
    s, id = py_scanstring(s, end, strict, **kwargs)
    return (slice(end, id), s), id


def quote(object, *, quotes="'''"):
    if quotes in object:
        quotes = '"""'
    return quotes + object + "\n" + quotes


def object_pairs_hook(object) -> (slice, str):
    object = dict(object)
    if "cells" in object:
        return object["cells"]

    if "cell_type" in object:
        _, object["cell_type"] = object["cell_type"]

    for key in ["text", "source"]:
        if key in object:
            if object[key]:
                return (
                    slice(object[key][0][0].start, object[key][-1][0].stop),
                    object,
                    "".join(_[1] for _ in object[key]),
                )
    return slice(None), None, None


class LineCacheNotebookDecoder(JSONDecoder):
    def __init__(
        self,
        markdown=quote,
        code=textwrap.dedent,
        raw=partial(textwrap.indent, prefix="# "),
        **kwargs
    ):
        super().__init__(**kwargs)

        for key in ("markdown", "code", "raw"):
            setattr(self, "transform_" + key, locals().get(key))

        self.parse_string = scanstring
        self.object_pairs_hook = object_pairs_hook
        self.scan_once = py_make_scanner(self)

    def decode(self, object, filename):
        lines = []

        linecache.updatecache(filename)
        if filename in linecache.cache:
            linecache.cache[filename] = (
                linecache.cache[filename][0],
                linecache.cache[filename][1],
                lines,
                filename,
            )
        last, new, old = slice(0, 0), 0, 0
        for current, cell, source in super().decode(object):
            if cell:
                lines += ["\n"] * (
                    object[last.stop : current.start].splitlines().__len__()
                    - 1
                    + (old - new)
                )

                source = getattr(self, "transform_" + cell["cell_type"])(source)

                lines += list(map("{}\n".format, source.splitlines()))
                new, old = map(len, map(str.splitlines, (source, object[current])))
                if not lines[-1]:
                    lines.pop()
                last = current

        return "".join(lines)



# coding: utf-8
import ast
import json
import os
import sys
from pathlib import Path

from IPython import get_ipython, paths
from IPython.core import profiledir


def get_config(profile="default"):
    profile_dir = profiledir.ProfileDir()
    try:
        profile = profile_dir.find_profile_dir_by_name(paths.get_ipython_dir(), profile)
    except profiledir.ProfileDirError:
        os.makedirs(paths.get_ipython_dir(), exist_ok=True)
        profile = profile_dir.create_profile_dir_by_name(
            paths.get_ipython_dir(), profile
        )
    return Path(profile.location, "ipython_config.json")


def load_config():
    location = get_config()
    try:
        with location.open() as file:
            config = json.load(file)
    except (FileNotFoundError, getattr(json, "JSONDecodeError", ValueError)):
        config = {}

    if "InteractiveShellApp" not in config:
        config["InteractiveShellApp"] = {}

    if "extensions" not in config["InteractiveShellApp"]:
        config["InteractiveShellApp"]["extensions"] = []

    return config, location


def install(project="importnb"):
    config, location = load_config()
    projects = sys.argv[1:] or [project]
    if not installed(project):
        config["InteractiveShellApp"]["extensions"].extend(projects)

    with location.open("w") as file:
        json.dump(config, file)

    print("""<3 {}""".format(projects))


def installed(project):
    config, location = load_config()
    return project in config.get("InteractiveShellApp", {}).get("extensions", [])


def uninstall(project="importnb"):
    config, location = load_config()
    projects = sys.argv[1:] or [project]
    config["InteractiveShellApp"]["extensions"] = [
        ext
        for ext in config["InteractiveShellApp"]["extensions"]
        if ext not in projects
    ]

    with location.open("w") as file:
        json.dump(config, file)
    print("""</3 {}.""".format(projects))


if __name__ == "__main__":
    from importnb.utils.export import export

    export("ipython.ipynb", "../../utils/ipython.py")



# coding: utf-8
"""# `loader`

Combine the __import__ finder with the loader.
"""


import ast
import importlib
import inspect
import json
import os
import sys
import textwrap
import types
from contextlib import ExitStack, contextmanager
from functools import partial, partialmethod
from importlib import reload
from importlib.machinery import ModuleSpec, SourceFileLoader
from importlib.util import spec_from_loader
from inspect import signature
from pathlib import Path

from .decoder import LineCacheNotebookDecoder, quote
from .docstrings import update_docstring
from .finder import FuzzyFinder, FuzzySpec, get_loader_details
from .ipython_extension import load_ipython_extension, unload_ipython_extension

_GTE38 = sys.version_info.major == 3 and sys.version_info.minor >= 8

if _GTE38:
    from importlib._bootstrap import _load_unlocked, _requires_builtin
else:
    from importlib._bootstrap import _installed_safely, _requires_builtin


try:
    from importlib._bootstrap_external import decode_source, FileFinder
    from importlib.util import module_from_spec
    from importlib._bootstrap import _init_module_attrs
    from importlib.util import LazyLoader
except ImportError:
    # python 3.4
    from importlib._bootstrap import _SpecMethods
    from importlib.util import decode_source
    from importlib.machinery import FileFinder

    def module_from_spec(spec):
        return _SpecMethods(spec).create()

    def _init_module_attrs(spec, module):
        return _SpecMethods(spec).init_module_attrs(module)


try:
    import IPython
    from IPython.core.inputsplitter import IPythonInputSplitter

    dedent = IPythonInputSplitter(
        line_input_checker=False,
        physical_line_transforms=[
            IPython.core.inputsplitter.leading_indent(),
            IPython.core.inputsplitter.ipy_prompt(),
            IPython.core.inputsplitter.cellmagic(end_on_blank_line=False),
        ],
    ).transform_cell
except:
    from textwrap import dedent

__all__ = "Notebook", "reload"


class FinderContextManager:
    """
    FinderContextManager is the base class for the notebook loader.  It provides
    a context manager that replaces `FileFinder` in the `sys.path_hooks` to include
    an instance of the class in the python findering system.

    >>> with FinderContextManager() as f:
    ...      id, ((loader_cls, _), *_) = get_loader_details()
    ...      assert issubclass(loader_cls, FinderContextManager)
    >>> id, ((loader_cls, _), *_) = get_loader_details()
    >>> loader_cls = inspect.unwrap(loader_cls)
    >>> assert not (isinstance(loader_cls, type) and issubclass(loader_cls, FinderContextManager))
    """

    extensions = tuple()
    _position = 0

    finder = FileFinder

    @property
    def loader(self):
        return type(self)

    def __enter__(self):
        id, details = get_loader_details()
        details.insert(self._position, (self.loader, self.extensions))
        sys.path_hooks[id] = self.finder.path_hook(*details)
        sys.path_importer_cache.clear()
        return self

    def __exit__(self, *excepts):
        id, details = get_loader_details()
        details.pop(self._position)
        sys.path_hooks[id] = self.finder.path_hook(*details)
        sys.path_importer_cache.clear()


"""## The basic loader

The loader uses the import systems `get_source`, `get_data`, and `create_module` methods to import notebook files.
"""


class ModuleType(types.ModuleType, getattr(os, "PathLike", object)):
    """ModuleType combines a module with a PathLike access to simplify access."""

    def __fspath__(self):
        return self.__file__


class ImportLibMixin(SourceFileLoader):
    """ImportLibMixin is a SourceFileLoader for loading source code from JSON (e.g. notebooks).

    `get_data` assures consistent line numbers between the file s representatio and source."""

    def create_module(self, spec):
        module = ModuleType(spec.name)
        _init_module_attrs(spec, module)
        if isinstance(spec, FuzzySpec):
            sys.modules[spec.alias] = module
        if self.name:
            module.__name__ = self.name
        return module

    def decode(self):
        return decode_source(super().get_data(self.path))

    def get_data(self, path):
        """Needs to return the string source for the module."""
        return LineCacheNotebookDecoder(
            code=self.code, raw=self.raw, markdown=self.markdown
        ).decode(self.decode(), self.path)

    @classmethod
    @_requires_builtin
    def is_package(cls, fullname):
        """Return False as built-in modules are never packages."""
        if "." not in fullname:
            return True
        return super().is_package(fullname)

    get_source = get_data


class NotebookBaseLoader(ImportLibMixin, FinderContextManager):
    """The simplest implementation of a Notebook Source File Loader.
    """

    extensions = (".ipynb",)
    __slots__ = "_lazy", "_fuzzy", "_markdown_docstring", "_position"

    def __init__(
        self,
        fullname=None,
        path=None,
        *,
        lazy=False,
        fuzzy=True,
        position=0,
        markdown_docstring=True
    ):
        super().__init__(fullname, path)
        self._lazy = lazy
        self._fuzzy = fuzzy
        self._markdown_docstring = markdown_docstring
        self._position = position

    @property
    def loader(self):
        """Create a lazy loader source file loader."""
        loader = super().loader
        if self._lazy and (sys.version_info.major, sys.version_info.minor) != (3, 4):
            loader = LazyLoader.factory(loader)
        # Strip the leading underscore from slots
        return partial(
            loader,
            **{object.lstrip("_"): getattr(self, object) for object in self.__slots__}
        )

    @property
    def finder(self):
        """Permit fuzzy finding of files with special characters."""
        return self._fuzzy and FuzzyFinder or super().finder


class FileModuleSpec(ModuleSpec):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._set_fileattr = True


"""## notebooks most be loadable from files.
"""


class FromFileMixin:
    """FromFileMixin adds a classmethod to load a notebooks from files.
    """

    @classmethod
    def load(cls, filename, dir=None, main=False, **kwargs):
        """Import a notebook as a module from a filename.

        dir: The directory to load the file from.
        main: Load the module in the __main__ context.

        > assert Notebook.load('loader.ipynb')
        """
        name = main and "__main__" or Path(filename).stem
        loader = cls(name, str(filename), **kwargs)
        spec = FileModuleSpec(name, loader, origin=loader.path)
        module = module_from_spec(spec)
        if _GTE38:
            module = _load_unlocked(spec)
        else:
            with ExitStack() as stack:
                loader.name != "__main__" and stack.enter_context(
                    _installed_safely(module)
                )
                loader.exec_module(module)
        return module


"""* Sometimes folks may want to use the current IPython shell to manage the code and input transformations.
"""

"""Use the `IPythonInputSplitter` to dedent and process magic functions.
"""


class TransformerMixin:
    def code(self, str):
        return dedent(str)

    def markdown(self, str):
        return quote(str)

    def raw(self, str):
        return textwrap.indent(str, "# ")

    def visit(self, node):
        return node


"""## The `Notebook` finder & loader
"""


class Notebook(TransformerMixin, FromFileMixin, NotebookBaseLoader):
    """Notebook is a user friendly file finder and module loader for notebook source code.

    > Remember, restart and run all or it didn't happen.

    Notebook provides several useful options.

    * Lazy module loading.  A module is executed the first time it is used in a script.
    """

    __slots__ = NotebookBaseLoader.__slots__ + ("_main",)

    def __init__(
        self,
        fullname=None,
        path=None,
        lazy=False,
        position=0,
        fuzzy=True,
        markdown_docstring=True,
        main=False,
    ):
        self._main = bool(main) or fullname == "__main__"
        super().__init__(
            self._main and "__main__" or fullname,
            path,
            lazy=lazy,
            fuzzy=fuzzy,
            position=position,
            markdown_docstring=markdown_docstring,
        )

    def parse(self, nodes):
        return ast.parse(nodes, self.path)

    def source_to_code(self, nodes, path, *, _optimize=-1):
        """* Convert the current source to ast
        * Apply ast transformers.
        * Compile the code."""
        if not isinstance(nodes, ast.Module):
            nodes = self.parse(nodes)
        if self._markdown_docstring:
            nodes = update_docstring(nodes)
        return super().source_to_code(
            ast.fix_missing_locations(self.visit(nodes)), path, _optimize=_optimize
        )


"""# Developer
"""

"""    Notebook.load('loader.ipynb')

"""

if __name__ == "__main__":
    try:
        from utils.export import export
    except:
        from .utils.export import export
    export("loader.ipynb", "../loader.py")
    print(__import__("doctest").testmod(Notebook.load("loader.ipynb"), verbose=2))



# coding: utf-8
"""# Special handling of markdown cells as docstrings.

Modify the Python `ast` to assign docstrings to functions when they are preceded by a Markdown cell.
"""

import ast

"""# Modifying the `ast`

    >>> assert isinstance(create_test, ast.Assign)
    >>> assert isinstance(test_update, ast.Attribute)
"""

create_test = ast.parse(
    """__test__ = globals().get('__test__', {})""", mode="single"
).body[0]
test_update = ast.parse("""__test__.update""", mode="single").body[0].value
str_nodes = (ast.Str,)

"""`TestStrings` is an `ast.NodeTransformer` that captures `str_nodes` in the `TestStrings.strings` object.

```ipython
>>> assert isinstance(ast.parse(TestStrings().visit(ast.parse('"Test me"'))), ast.Module)

```
"""


class TestStrings(ast.NodeTransformer):

    strings = None

    def visit_Module(self, module):
        """`TestStrings.visit_Module` initializes the capture. After all the nodes are visit we append `create_test and test_update`
        to populate the `"__test__"` attribute.
        """
        self.strings = []
        module = self.visit_body(module)
        module.body += (
            [create_test]
            + [
                ast.copy_location(
                    ast.Expr(
                        ast.Call(
                            func=test_update,
                            args=[
                                ast.Dict(
                                    keys=[ast.Str("string-{}".format(node.lineno))],
                                    values=[node],
                                )
                            ],
                            keywords=[],
                        )
                    ),
                    node,
                )
                for node in self.strings
            ]
            if self.strings
            else []
        )
        return module

    def visit_body(self, node):
        """`TestStrings.visit_body` visits nodes with a `"body"` attibute and extracts potential string tests."""

        body = []
        if (
            node.body
            and isinstance(node.body[0], ast.Expr)
            and isinstance(node.body[0].value, str_nodes)
        ):
            body.append(node.body.pop(0))
        node.body = body + [
            (self.visit_body if hasattr(object, "body") else self.visit)(object)
            for object in node.body
        ]
        return node

    def visit_Expr(self, node):
        """`TestStrings.visit_Expr` append the `str_nodes` to `TestStrings.strings` to append to the `ast.Module`."""

        if isinstance(node.value, str_nodes):
            self.strings.append(
                ast.copy_location(ast.Str(node.value.s.replace("\n```", "\n")), node)
            )
        return node


def update_docstring(module):
    from functools import reduce

    module.body = reduce(markdown_docstring, module.body, [])
    return TestStrings().visit(module)


docstring_ast_types = ast.ClassDef, ast.FunctionDef
try:
    docstring_ast_types += (ast.AsyncFunctionDef,)
except:
    ...


def markdown_docstring(nodes, node):
    if (
        len(nodes) > 1
        and str_expr(nodes[-1])
        and isinstance(node, docstring_ast_types)
        and not str_expr(node.body[0])
    ):
        node.body.insert(0, nodes.pop())
    return nodes.append(node) or nodes


def str_expr(node):
    return isinstance(node, ast.Expr) and isinstance(node.value, ast.Str)



# coding: utf-8
"""A `pytest` plugin for importing notebooks as modules and using standard test discovered.

The `AlternativeModule` is reusable.  See `pidgin` for an example.
"""

import abc
import functools
import importlib
import pathlib

import _pytest
import pytest

from importnb import Notebook


def pytest_addoption(parser):
    group = parser.getgroup("general")
    group.addoption("--main", action="store_true", help="Run in the main context.")


"""`AlternativeModule` is an alternative `pytest.Module` loader that can enable `pytest.Doctest`.
"""


class AlternativeModule(pytest.Module):
    def _getobj(self):
        return self.loader(
            getattr(self.parent.config.option, "main", None) and "__main__" or None
        ).load(str(self.fspath))

    def collect(self):
        yield from super().collect()
        if self.parent.config.option.doctestmodules:
            self.fspath.pyimport = functools.partial(
                self.fspath.pyimport, modname=self._obj.__name__
            )
            yield from _pytest.doctest.DoctestModule.collect(self)


"""`NotebookModule` is an `AlternativeModule` to load `Notebook`s.
"""


class NotebookModule(AlternativeModule):
    loader = Notebook


class AlternativeSourceText(abc.ABCMeta):
    def __call__(self, parent, path):
        for module in self.modules:
            if "".join(pathlib.Path(str(path)).suffixes) in module.loader.extensions:
                if not parent.session.isinitpath(path):
                    for pat in parent.config.getini("python_files"):
                        if path.fnmatch(pat.rstrip(".py") + path.ext):
                            break
                    else:
                        return
                return module(path, parent)


class NotebookTests(metaclass=AlternativeSourceText):
    modules = (NotebookModule,)


pytest_collect_file = NotebookTests.__call__

if __name__ == "__main__":
    from importnb.utils.export import export

    export("pytest_importnb.ipynb", "../../utils/pytest_importnb.py")



# coding: utf-8

__all__ = "Notebook", "reload", "Parameterize"

from ._version import *
from .ipython_extension import load_ipython_extension, unload_ipython_extension
from .loader import Notebook, reload, unload_ipython_extension



# coding: utf-8
"""It is important to distribute notebooks in packages during the initial stages of code development.  This notebook creates a setuptools command class that allows for both python and notebook imports.  This was specifically created to allow notebooks as py_module imports, but could serve a greater purpose.
"""

"""    class BuildWithNotebooks(setuptools.command.build_py.build_py):
        def __new__(cls, distribution):
            from importnb.utils.setup import build_ipynb
            return build_ipynb(distribution)
    setup_args.update(cmdclass=dict(build_py=BuildWithNotebooks))

"""

import importlib
import os
import sys
from pathlib import Path

from setuptools.command.build_py import build_py


class build_ipynb(build_py):
    """Should really use manifest.in

    Lazy import build_ipynb in your setup.

    class BuildWithNotebooks(setuptools.command.build_py.build_py):
        def __new__(cls, distribution):
            from importnb.utils.setup import build_ipynb
            return build_ipynb(distribution)
    setup_args.update(cmdclass=dict(build_py=BuildWithNotebooks))
    """

    def get_module_outfile(self, build_dir, package, module):
        module_mapper = {module[1]: module[2] for module in self.find_all_modules()}
        outfile_path = [build_dir] + list(package) + [module_mapper[module]]
        return os.path.join(*outfile_path)

    def find_package_modules(self, package, package_dir):
        from glob import glob

        self.check_package(package, package_dir)
        module_files = glob(os.path.join(package_dir, "*.py"))
        modules = []
        setup_script = os.path.abspath(self.distribution.script_name)

        for f in module_files + glob(os.path.join(package_dir, "*.ipynb")):
            abs_f = os.path.abspath(f)
            if abs_f != setup_script:
                module = os.path.splitext(os.path.basename(f))[0]
                modules.append((package, module, f))
            else:
                self.debug_print("excluding %s" % setup_script)
        return modules

    def find_modules(self):
        packages, modules = {}, []

        for module in self.py_modules:
            path = module.split(".")
            package = ".".join(path[0:-1])
            module_base = path[-1]

            try:
                (package_dir, checked) = packages[package]
            except KeyError:
                package_dir = self.get_package_dir(package)
                checked = 0

            if not checked:
                init_py = self.check_package(package, package_dir)
                packages[package] = (package_dir, 1)
                if init_py:
                    modules.append((package, "__init__", init_py))

            module_file = os.path.join(package_dir, module_base + ".ipynb")

            if Path(module_file).exists():
                modules.append((package, module_base, str(module_file)))
            else:
                module_file = str(Path(module_file).with_suffix(".py"))
                if self.check_module(module, module_file):
                    modules.append((package, module_base, str(module_file)))

        return modules



# coding: utf-8
"""# `sys.path_hook` modifiers

Many suggestions for importing notebooks use `sys.meta_paths`, but `importnb` relies on the `sys.path_hooks` to load any notebook in the path. `PathHooksContext` is a base class for the `importnb.Notebook` `SourceFileLoader`.
"""

import ast
import inspect
import os
import sys
from contextlib import ExitStack, contextmanager
from importlib.machinery import ModuleSpec, SourceFileLoader
from itertools import chain
from pathlib import Path

try:
    from importlib._bootstrap_external import FileFinder
except ImportError:
    # python 3.4
    from importlib.machinery import FileFinder


class FileModuleSpec(ModuleSpec):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._set_fileattr = True


class FuzzySpec(FileModuleSpec):
    def __init__(
        self,
        name,
        loader,
        *,
        alias=None,
        origin=None,
        loader_state=None,
        is_package=None
    ):
        super().__init__(
            name,
            loader,
            origin=origin,
            loader_state=loader_state,
            is_package=is_package,
        )
        self.alias = alias


def fuzzy_query(str):
    new = ""
    for chr in str:
        new += (not new.endswith("__") or chr != "_") and chr or ""
    return new.replace("__", "*").replace("_", "?")


def fuzzy_file_search(path, fullname):
    results = []
    id, details = get_loader_details()
    for ext in sum((list(object[1]) for object in details), []):
        results.extend(Path(path).glob(fullname + ext))
        "_" in fullname and results.extend(Path(path).glob(fuzzy_query(fullname) + ext))
    return results


class FuzzyFinder(FileFinder):
    """Adds the ability to open file names with special characters using underscores."""

    def find_spec(self, fullname, target=None):
        """Try to finder the spec and if it cannot be found, use the underscore starring syntax
        to identify potential matches.
        """
        spec = super().find_spec(fullname, target=target)

        if spec is None:
            original = fullname

            if "." in fullname:
                original, fullname = fullname.rsplit(".", 1)
            else:
                original, fullname = "", original

            if "_" in fullname:
                files = fuzzy_file_search(self.path, fullname)
                if files:
                    file = Path(sorted(files)[0])
                    spec = super().find_spec(
                        (original + "." + file.stem.split(".", 1)[0]).lstrip("."),
                        target=target,
                    )
                    fullname = (original + "." + fullname).lstrip(".")
                    if spec and fullname != spec.name:
                        spec = FuzzySpec(
                            spec.name,
                            spec.loader,
                            origin=spec.origin,
                            loader_state=spec.loader_state,
                            alias=fullname,
                            is_package=bool(spec.submodule_search_locations),
                        )
        return spec


def get_loader_details():
    for id, path_hook in enumerate(sys.path_hooks):
        try:
            return (
                id,
                list(inspect.getclosurevars(path_hook).nonlocals["loader_details"]),
            )
        except:
            continue



# coding: utf-8
"""# The `export` module

...provides compatibility for Python and IPython through [`compile_python`](compile_python.ipynb) and [`compile_ipython`](compile_ipython.ipynb), respectively.

    >>> from importnb.utils.export import export
"""

from json import loads
from pathlib import Path

from ..loader import dedent

try:
    from black import format_str
except ImportError:
    format_str = lambda x, i: x


def block_str(str):
    quotes = '"""'
    if quotes in str:
        quotes = "'''"
    return "{quotes}{str}\n{quotes}\n".format(quotes=quotes, str=str)


"""The export function
"""


def export(file, to=None):
    code = """# coding: utf-8"""
    with open(str(file), "r") as f:
        for cell in loads(f.read())["cells"]:
            if cell["cell_type"] == "markdown":
                code += "\n" + block_str("".join(cell["source"]))
            elif cell["cell_type"] == "code":
                code += "\n" + dedent("".join(cell["source"]))
    to and Path(to).with_suffix(".py").write_text(format_str(code, 100))
    return code



# coding: utf-8
"""# Fuzzy completion

The fuzzy importer could be confusing and perhaps a completer could help.


    >>> ip = __import__("IPython").get_ipython(); load_ipython_extension(ip)
    >>> assert ip.complete('importnb.__pleter', 'import importnb.__pleter')[1]
    >>> assert ip.complete('__find__', 'import __find__')[1]
    >>> assert ip.complete('IPy', '\timport IPy')[1]
    >>> assert ip.complete('_______________plet__', 'from importnb import _______________plet__')[1]
"""

import string
from fnmatch import fnmatch
from pathlib import Path

from .finder import fuzzy_file_search


"""To provide the most reliable fuzzy imports `fuzzify_string` replaces the imported with one that complies with the fuzzy finder.
"""


def fuzzify_string(str):
    return (str[0] in string.ascii_letters + "_" and str[0] or "_") + "".join(
        letter if letter in string.ascii_letters + "_" + string.digits else "_"
        for letter in str[1:]
    )


"""`align_match` replaces the the begining of the match with a prefix that matches that completer query name.
"""


def align_match(match, prefix, *, i=0):
    pattern = prefix.replace("__", "*").replace("_", "?").strip()
    for i in range(len(match)):
        if fnmatch(match[:i], pattern):
            break
    else:
        i += 1
    return prefix + match[i:]


"""* `predict_fuzzy` will take a fully qualified fuzzy name completions.  This is the main function for the completer.
"""


def predict_fuzzy(fullname):
    package, paths, specs, extras = "", [], [], []
    if "." in fullname:
        package, fullname = fullname.rsplit(".", 1)
        fullname = fullname.strip()
        try:
            module = __import__("importlib").import_module(package)
            paths.append(Path(module.__file__).parent)
            extras = [object for object in dir(module) if object.startswith("fullname")]
        except:
            ...
    else:
        paths = map(Path, __import__("sys").path)
    query = fullname
    while not query.endswith("__"):
        query += "_"
    for path in paths:
        specs.extend(
            str(object.relative_to(path).with_suffix(""))
            for object in fuzzy_file_search(path, query)
        )

    return set(
        (package and package + "." or "") + align_match(fuzzify_string(spec), fullname)
        for spec in specs
    ).union(set(extras))


def fuzzy_complete_event(self, event):
    event.line = event.line.lstrip()
    symbol = event.symbol
    if event.line.startswith("from"):
        package = event.line.split(" import ", 1)[0].lstrip().lstrip("from").lstrip()
        if " import" in event.line:
            symbol = (package + "." + symbol).lstrip(".")
            return [
                object.lstrip(package).lstrip(".") for object in predict_fuzzy(symbol)
            ]

    return predict_fuzzy(symbol)


"""* The extension adds the new fuzzy completer.  Our completer has a higher priority than the default completers.  Since we stripped the leading whitespace from the completion line event; the extension will permit completion on tabbed lines.
"""


def load_ipython_extension(ip):
    ip.set_hook(
        "complete_command", fuzzy_complete_event, str_key="aimport", priority=25
    )
    ip.set_hook("complete_command", fuzzy_complete_event, str_key="import", priority=25)
    ip.set_hook(
        "complete_command", fuzzy_complete_event, str_key="%reload_ext", priority=25
    )
    ip.set_hook(
        "complete_command", fuzzy_complete_event, str_key="%load_ext", priority=25
    )
    ip.set_hook("complete_command", fuzzy_complete_event, str_key="from", priority=25)
